from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class HDF5(DefinedNamespace):
    # Generated with ontolutils version 0.13.2
    _fail = True
    AllSelection: URIRef
    AndSelection: URIRef
    ArrayDatatype: URIRef
    ArrayDimension: URIRef
    ArrayValue: URIRef
    AtomicDatatype: URIRef
    Attribute: URIRef
    BitfieldDatatype: URIRef
    CharacterSet: URIRef
    Chunk: URIRef
    ChunkDimension: URIRef
    ChunkMode: URIRef
    CompositeDatatype: URIRef
    CompositeValue: URIRef
    CompoundDatatype: URIRef
    CompoundDatatypeMember: URIRef
    CompoundValue: URIRef
    Coordinate: URIRef
    DataStream: URIRef
    Dataset: URIRef
    Dataspace: URIRef
    DataspaceDimension: URIRef
    Datatype: URIRef
    Dimension: URIRef
    DimensionSelection: URIRef
    EnumerationDatatype: URIRef
    EnumerationDatatypeMember: URIRef
    ErrorDetectionCheck: URIRef
    ExponentField: URIRef
    ExternalFile: URIRef
    FieldSpecification: URIRef
    File: URIRef
    FileAccessMode: URIRef
    FillTime: URIRef
    FillValueStatus: URIRef
    Filter: URIRef
    FilterDefault: URIRef
    FilterDeflate: URIRef
    FilterFletcher: URIRef
    FilterNBit: URIRef
    FilterSZip: URIRef
    FilterScaleOffset: URIRef
    FilterShuffle: URIRef
    FloatAttribute: URIRef
    FloatDatatype: URIRef
    FloatingPointNormalization: URIRef
    Group: URIRef
    HardLink: URIRef
    HyperSlab: URIRef
    IntegerAttribute: URIRef
    IntegerDatatype: URIRef
    Link: URIRef
    MantissaField: URIRef
    MessageTypeFlag: URIRef
    NameValuePair: URIRef
    NamedDatatype: URIRef
    NamedObject: URIRef
    NoneSelection: URIRef
    NotASelection: URIRef
    NotBSelection: URIRef
    OpaqueAttribute: URIRef
    OpaqueDatatype: URIRef
    OrSelection: URIRef
    Order: URIRef
    Padding: URIRef
    Point: URIRef
    PointSelection: URIRef
    PointValuePair: URIRef
    ReferenceDatatype: URIRef
    SZipCoding: URIRef
    ScalarDataspace: URIRef
    ScaleType: URIRef
    Selection: URIRef
    Sign: URIRef
    SignField: URIRef
    SimpleDataspace: URIRef
    SoftLink: URIRef
    StorageAllocation: URIRef
    StorageLayout: URIRef
    StringAttribute: URIRef
    StringDatatype: URIRef
    StringPadding: URIRef
    TimeDatatype: URIRef
    TransferMode: URIRef
    TypeClass: URIRef
    VariableLengthDatatype: URIRef
    XOrSelection: URIRef
    allocationTime: URIRef
    attribute: URIRef
    characterSet: URIRef
    chunk: URIRef
    chunkMode: URIRef
    compositeFillValue: URIRef
    compositeValue: URIRef
    coordinate: URIRef
    dataSource: URIRef
    dataTarget: URIRef
    dataspace: URIRef
    datatype: URIRef
    dimension: URIRef
    dimensionSelection: URIRef
    errorDetectionCheck: URIRef
    exponentField: URIRef
    externalLinkAccess: URIRef
    field: URIRef
    fillTime: URIRef
    fillValueStatus: URIRef
    filter: URIRef
    internalPadding: URIRef
    layout: URIRef
    link: URIRef
    linkedTo: URIRef
    lsbPadding: URIRef
    mantissaField: URIRef
    member: URIRef
    msbPadding: URIRef
    multiDimensionalIndex: URIRef
    normalization: URIRef
    notA: URIRef
    notB: URIRef
    order: URIRef
    point: URIRef
    previous: URIRef
    rootGroup: URIRef
    scaleType: URIRef
    selectionOn: URIRef
    sign: URIRef
    signField: URIRef
    stringPadding: URIRef
    szipEncoding: URIRef
    transferMode: URIRef
    typeClass: URIRef
    writtenRegion: URIRef
    xor: URIRef
    atomicFillValue: URIRef
    block: URIRef
    btreeIStoreSize: URIRef
    btreeLeaf: URIRef
    btreeRank: URIRef
    btreeRatioLeft: URIRef
    btreeRatioMiddle: URIRef
    btreeRatioRight: URIRef
    bufferSize: URIRef
    chunkCachePreemptionPolicy: URIRef
    chunkCacheSize: URIRef
    chunkCacheSlots: URIRef
    collectiveIOThreshold: URIRef
    compressionLevel: URIRef
    coordinateIndex: URIRef
    count: URIRef
    createIntermediateGroups: URIRef
    currentSize: URIRef
    data: URIRef
    deflateLevel: URIRef
    dimensionIndex: URIRef
    ebias: URIRef
    externalLinkCacheSize: URIRef
    externalLinkPrefix: URIRef
    globalFreelistVersion: URIRef
    hyperVectorSize: URIRef
    iStoreK: URIRef
    index: URIRef
    index_0: URIRef
    index_1: URIRef
    index_10: URIRef
    index_11: URIRef
    index_12: URIRef
    index_13: URIRef
    index_14: URIRef
    index_15: URIRef
    index_16: URIRef
    index_17: URIRef
    index_18: URIRef
    index_19: URIRef
    index_2: URIRef
    index_20: URIRef
    index_21: URIRef
    index_22: URIRef
    index_23: URIRef
    index_24: URIRef
    index_25: URIRef
    index_26: URIRef
    index_27: URIRef
    index_28: URIRef
    index_29: URIRef
    index_3: URIRef
    index_30: URIRef
    index_31: URIRef
    index_4: URIRef
    index_5: URIRef
    index_6: URIRef
    index_7: URIRef
    index_8: URIRef
    index_9: URIRef
    initialSize: URIRef
    isVariableLength: URIRef
    linkedChunkThreshold: URIRef
    maximumLinkTraversals: URIRef
    maximumSize: URIRef
    metaBlockSize: URIRef
    minimumSize: URIRef
    name: URIRef
    offset: URIRef
    pixelsPerBlock: URIRef
    position: URIRef
    precision: URIRef
    preserve: URIRef
    rank: URIRef
    scaleFactor: URIRef
    sharedObjectHeaderVersion: URIRef
    sieveBufferSize: URIRef
    size: URIRef
    sizeOfAddress: URIRef
    sizeOfSize: URIRef
    smallDataBlockSize: URIRef
    start: URIRef
    stride: URIRef
    superBlockVersion: URIRef
    symbolTableNodeSize: URIRef
    symbolTableTreeRank: URIRef
    symbolTableVersion: URIRef
    trackTimes: URIRef
    userBlockSize: URIRef
    value: URIRef
    H5D_ALLOC_TIME_DEFAULT: URIRef
    H5D_ALLOC_TIME_EARLY: URIRef
    H5D_ALLOC_TIME_INCR: URIRef
    H5D_ALLOC_TIME_LATE: URIRef
    H5D_CHUNKED: URIRef
    H5D_COMPACT: URIRef
    H5D_CONTIGUOUS: URIRef
    H5D_FILL_TIME_ALLOC: URIRef
    H5D_FILL_TIME_IFSET: URIRef
    H5D_FILL_TIME_NEVER: URIRef
    H5D_FILL_VALUE_DEFAULT: URIRef
    H5D_FILL_VALUE_UNDEFINED: URIRef
    H5D_FILL_VALUE_USER_DEFINED: URIRef
    H5D_VIRTUAL: URIRef
    H5FD_MPIO_CHUNK_MULTI_IO: URIRef
    H5FD_MPIO_CHUNK_ONE_IO: URIRef
    H5FD_MPIO_COLLECTIVE: URIRef
    H5FD_MPIO_INDEPENDENT: URIRef
    H5F_ACC_DEFAULT: URIRef
    H5F_ACC_RDONLY: URIRef
    H5F_ACC_RDWR: URIRef
    H5O_SHMESG_ALL_FLAG: URIRef
    H5O_SHMESG_ATTR_FLAG: URIRef
    H5O_SHMESG_DTYPE_FLAG: URIRef
    H5O_SHMESG_FILL_FLAG: URIRef
    H5O_SHMESG_NONE_FLAG: URIRef
    H5O_SHMESG_PLINE_FLAG: URIRef
    H5O_SHMESG_SDSPACE_FLAG: URIRef
    H5P_DATASET_CREATE: URIRef
    H5P_DATASET_XFER: URIRef
    H5P_FILE_ACCESS: URIRef
    H5P_FILE_CREATE: URIRef
    H5P_FILE_MOUNT: URIRef
    H5T_ALPHA_B16: URIRef
    H5T_ALPHA_B32: URIRef
    H5T_ALPHA_B64: URIRef
    H5T_ALPHA_B8: URIRef
    H5T_ALPHA_F32: URIRef
    H5T_ALPHA_F64: URIRef
    H5T_ALPHA_I16: URIRef
    H5T_ALPHA_I32: URIRef
    H5T_ALPHA_I64: URIRef
    H5T_ALPHA_I8: URIRef
    H5T_ALPHA_U16: URIRef
    H5T_ALPHA_U32: URIRef
    H5T_ALPHA_U64: URIRef
    H5T_ALPHA_U8: URIRef
    H5T_ARRAY: URIRef
    H5T_BITFIELD: URIRef
    H5T_COMPOUND: URIRef
    H5T_CSET_ASCII: URIRef
    H5T_CSET_UTF8: URIRef
    H5T_ENUM: URIRef
    H5T_FLOAT: URIRef
    H5T_IEEE_F32BE: URIRef
    H5T_IEEE_F32LE: URIRef
    H5T_IEEE_F64BE: URIRef
    H5T_IEEE_F64LE: URIRef
    H5T_INTEGER: URIRef
    H5T_INTEL_B16: URIRef
    H5T_INTEL_B32: URIRef
    H5T_INTEL_B64: URIRef
    H5T_INTEL_B8: URIRef
    H5T_INTEL_F32: URIRef
    H5T_INTEL_F64: URIRef
    H5T_INTEL_I16: URIRef
    H5T_INTEL_I32: URIRef
    H5T_INTEL_I64: URIRef
    H5T_INTEL_I8: URIRef
    H5T_INTEL_U16: URIRef
    H5T_INTEL_U32: URIRef
    H5T_INTEL_U64: URIRef
    H5T_INTEL_U8: URIRef
    H5T_MIPS_B16: URIRef
    H5T_MIPS_B32: URIRef
    H5T_MIPS_B64: URIRef
    H5T_MIPS_B8: URIRef
    H5T_MIPS_F32: URIRef
    H5T_MIPS_F64: URIRef
    H5T_MIPS_I16: URIRef
    H5T_MIPS_I32: URIRef
    H5T_MIPS_I64: URIRef
    H5T_MIPS_I8: URIRef
    H5T_MIPS_U16: URIRef
    H5T_MIPS_U32: URIRef
    H5T_MIPS_U64: URIRef
    H5T_MIPS_U8: URIRef
    H5T_NATIVE_CHAR: URIRef
    H5T_NATIVE_DOUBLE: URIRef
    H5T_NATIVE_FLOAT: URIRef
    H5T_NATIVE_INT: URIRef
    H5T_NATIVE_LDOUBLE: URIRef
    H5T_NATIVE_LLONG: URIRef
    H5T_NATIVE_LONG: URIRef
    H5T_NATIVE_SHORT: URIRef
    H5T_NATIVE_UCHAR: URIRef
    H5T_NATIVE_UINT: URIRef
    H5T_NATIVE_ULLONG: URIRef
    H5T_NATIVE_ULONG: URIRef
    H5T_NATIVE_USHORT: URIRef
    H5T_NORM_IMPLIED: URIRef
    H5T_NORM_MSBSET: URIRef
    H5T_NORM_NONE: URIRef
    H5T_OPAQUE: URIRef
    H5T_ORDER_BE: URIRef
    H5T_ORDER_LE: URIRef
    H5T_ORDER_MIXED: URIRef
    H5T_ORDER_VAX: URIRef
    H5T_PAD_BACKGROUND: URIRef
    H5T_PAD_ONE: URIRef
    H5T_PAD_ZERO: URIRef
    H5T_REFERENCE: URIRef
    H5T_SGN_2: URIRef
    H5T_SGN_NONE: URIRef
    H5T_STD_B16BE: URIRef
    H5T_STD_B16LE: URIRef
    H5T_STD_B32BE: URIRef
    H5T_STD_B32LE: URIRef
    H5T_STD_B64BE: URIRef
    H5T_STD_B64LE: URIRef
    H5T_STD_B8BE: URIRef
    H5T_STD_B8LE: URIRef
    H5T_STD_I16BE: URIRef
    H5T_STD_I16LE: URIRef
    H5T_STD_I32BE: URIRef
    H5T_STD_I32LE: URIRef
    H5T_STD_I64BE: URIRef
    H5T_STD_I64LE: URIRef
    H5T_STD_I8BE: URIRef
    H5T_STD_I8LE: URIRef
    H5T_STD_U16BE: URIRef
    H5T_STD_U16LE: URIRef
    H5T_STD_U32BE: URIRef
    H5T_STD_U32LE: URIRef
    H5T_STD_U64BE: URIRef
    H5T_STD_U64LE: URIRef
    H5T_STD_U8BE: URIRef
    H5T_STD_U8LE: URIRef
    H5T_STRING: URIRef
    H5T_STR_NULLPAD: URIRef
    H5T_STR_NULLTERM: URIRef
    H5T_STR_SPACEPAD: URIRef
    H5T_TIME: URIRef
    H5T_VAX_F32: URIRef
    H5T_VAX_F64: URIRef
    H5T_VLEN: URIRef
    H5Z_DISABLE_EDC: URIRef
    H5Z_ENABLE_EDC: URIRef
    H5Z_SO_FLOAT_DSCALE: URIRef
    H5Z_SO_FLOAT_ESCALE: URIRef
    H5Z_SO_INT: URIRef
    H5_SZIP_EC_OPTION_MASK: URIRef
    H5_SZIP_NN_OPTION_MASK: URIRef
    scalarDataspace: URIRef

    _NS = Namespace("http://purl.allotrope.org/ontologies/hdf5/1.8#")