from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class SPDX(DefinedNamespace):
    _fail = True
    Annotation: URIRef
    AnnotationType: URIRef
    AnyLicenseInfo: URIRef
    Checksum: URIRef
    ChecksumAlgorithm: URIRef
    ConjunctiveLicenseSet: URIRef
    CreationInfo: URIRef
    CrossRef: URIRef
    DisjunctiveLicenseSet: URIRef
    ExternalDocumentRef: URIRef
    ExternalRef: URIRef
    ExtractedLicensingInfo: URIRef
    File: URIRef
    FileType: URIRef
    License: URIRef
    LicenseException: URIRef
    ListedLicense: URIRef
    ListedLicenseException: URIRef
    OrLaterOperator: URIRef
    Package: URIRef
    PackageVerificationCode: URIRef
    ReferenceCategory: URIRef
    ReferenceType: URIRef
    Relationship: URIRef
    RelationshipType: URIRef
    Review: URIRef
    SimpleLicensingInfo: URIRef
    Snippet: URIRef
    SpdxDocument: URIRef
    SpdxElement: URIRef
    SpdxItem: URIRef
    WithExceptionOperator: URIRef
    algorithm: URIRef
    annotation: URIRef
    annotationType: URIRef
    artifactOf: URIRef
    checksum: URIRef
    creationInfo: URIRef
    crossRef: URIRef
    dataLicense: URIRef
    describesPackage: URIRef
    externalDocumentRef: URIRef
    externalRef: URIRef
    fileDependency: URIRef
    fileType: URIRef
    hasExtractedLicensingInfo: URIRef
    hasFile: URIRef
    licenseConcluded: URIRef
    licenseDeclared: URIRef
    licenseException: URIRef
    licenseInfoFromFiles: URIRef
    licenseInfoInFile: URIRef
    licenseInfoInSnippet: URIRef
    member: URIRef
    packageVerificationCode: URIRef
    range: URIRef
    referenceCategory: URIRef
    referenceType: URIRef
    referencesFile: URIRef
    relatedSpdxElement: URIRef
    relationship: URIRef
    relationshipType: URIRef
    reviewed: URIRef
    snippetFromFile: URIRef
    spdxDocument: URIRef
    annotationDate: URIRef
    annotator: URIRef
    attributionText: URIRef
    checksumValue: URIRef
    contextualExample: URIRef
    copyrightText: URIRef
    created: URIRef
    creator: URIRef
    date: URIRef
    deprecatedVersion: URIRef
    description: URIRef
    documentation: URIRef
    downloadLocation: URIRef
    example: URIRef
    exceptionTextHtml: URIRef
    externalDocumentId: URIRef
    externalReferenceSite: URIRef
    extractedText: URIRef
    fileContributor: URIRef
    fileName: URIRef
    filesAnalyzed: URIRef
    isDeprecatedLicenseId: URIRef
    isFsfLibre: URIRef
    isLive: URIRef
    isOsiApproved: URIRef
    isValid: URIRef
    isWayBackLink: URIRef
    licenseComments: URIRef
    licenseExceptionId: URIRef
    licenseExceptionTemplate: URIRef
    licenseExceptionText: URIRef
    licenseId: URIRef
    licenseListVersion: URIRef
    licenseText: URIRef
    licenseTextHtml: URIRef
    match: URIRef
    name: URIRef
    noticeText: URIRef
    order: URIRef
    originator: URIRef
    packageFileName: URIRef
    packageName: URIRef
    packageVerificationCodeExcludedFile: URIRef
    packageVerificationCodeValue: URIRef
    referenceLocator: URIRef
    reviewDate: URIRef
    reviewer: URIRef
    snippetName: URIRef
    sourceInfo: URIRef
    specVersion: URIRef
    standardLicenseHeader: URIRef
    standardLicenseHeaderHtml: URIRef
    standardLicenseHeaderTemplate: URIRef
    standardLicenseTemplate: URIRef
    summary: URIRef
    supplier: URIRef
    timestamp: URIRef
    url: URIRef
    versionInfo: URIRef
    annotationType_other: URIRef
    annotationType_review: URIRef
    annotator: URIRef
    checksumAlgorithm_md2: URIRef
    checksumAlgorithm_md4: URIRef
    checksumAlgorithm_md5: URIRef
    checksumAlgorithm_md6: URIRef
    checksumAlgorithm_sha1: URIRef
    checksumAlgorithm_sha224: URIRef
    checksumAlgorithm_sha256: URIRef
    checksumAlgorithm_sha384: URIRef
    checksumAlgorithm_sha512: URIRef
    fileType_application: URIRef
    fileType_archive: URIRef
    fileType_audio: URIRef
    fileType_binary: URIRef
    fileType_documentation: URIRef
    fileType_image: URIRef
    fileType_other: URIRef
    fileType_source: URIRef
    fileType_spdx: URIRef
    fileType_text: URIRef
    fileType_video: URIRef
    noassertion: URIRef
    none: URIRef
    referenceCategory_other: URIRef
    referenceCategory_packageManager: URIRef
    referenceCategory_persistentId: URIRef
    referenceCategory_security: URIRef
    relationshipType_amendment: URIRef
    relationshipType_ancestorOf: URIRef
    relationshipType_buildDependencyOf: URIRef
    relationshipType_buildToolOf: URIRef
    relationshipType_containedBy: URIRef
    relationshipType_contains: URIRef
    relationshipType_copyOf: URIRef
    relationshipType_dataFile: URIRef
    relationshipType_dataFileOf: URIRef
    relationshipType_dependencyManifestOf: URIRef
    relationshipType_dependencyOf: URIRef
    relationshipType_dependsOn: URIRef
    relationshipType_descendantOf: URIRef
    relationshipType_describedBy: URIRef
    relationshipType_describes: URIRef
    relationshipType_devDependencyOf: URIRef
    relationshipType_devToolOf: URIRef
    relationshipType_distributionArtifact: URIRef
    relationshipType_documentation: URIRef
    relationshipType_dynamicLink: URIRef
    relationshipType_exampleOf: URIRef
    relationshipType_expandedFromArchive: URIRef
    relationshipType_fileAdded: URIRef
    relationshipType_fileDeleted: URIRef
    relationshipType_fileModified: URIRef
    relationshipType_generatedFrom: URIRef
    relationshipType_generates: URIRef
    relationshipType_hasPrerequisite: URIRef
    relationshipType_metafileOf: URIRef
    relationshipType_optionalComponentOf: URIRef
    relationshipType_optionalDependencyOf: URIRef
    relationshipType_other: URIRef
    relationshipType_packageOf: URIRef
    relationshipType_patchApplied: URIRef
    relationshipType_patchFor: URIRef
    relationshipType_prerequisiteFor: URIRef
    relationshipType_providedDependencyOf: URIRef
    relationshipType_runtimeDependencyOf: URIRef
    relationshipType_staticLink: URIRef
    relationshipType_testDependencyOf: URIRef
    relationshipType_testOf: URIRef
    relationshipType_testToolOf: URIRef
    relationshipType_testcaseOf: URIRef
    relationshipType_variantOf: URIRef
    reviewed: URIRef

    _NS = Namespace("http://spdx.org/rdf/terms#")